from .arcface import ArcFaceAdaptiveMarginLoss
from .circle import MultiLabelCircleLoss
from .colbert_loss import ColbertLoss
from .cosine_similarity import CosineSimilarity
from .infonce import InfoNCE
from .pearson_loss import PearsonLoss
from .simcse import SimCSE
from .token_loss import TokenLoss
from .triplet import TripletLoss
